//
//  Benchmarker.m
//  Benchmarker
//
//  Created by Tom Bernard on Mon May 17 2004.
//  Copyright (c) 2004 __MyCompanyName__. All rights reserved.
//

#import "Benchmarker.h"


@implementation Benchmarker

- (id)initBenchmarkerForSamples:(unsigned)_samples 
                     outputData:(BOOL)_outputData 
                  outputResults:(BOOL)_outputResults 
             withUnitMultiplier:(double)_multiplier
            disposeDataWhenDone:(BOOL)_disposeDataWhenDone
             nsLogStampStartEnd:(BOOL)_nsLogStampStartEnd
                      userLabel:(NSString *)_userLabel
{
    const double millisecondsMultiplier = 0.001;
    const double microsecondsMultiplier = 1.0;
    const double nanosecondsMultiplier = 1000.0;
    const double noResultsRequested = 0.0;

    if (self = [super init])
    {
        // allocate buffers for our data
        
        samples =               _samples;
        
        startMicrosecondsArray = calloc(samples,sizeof(UnsignedWide));
        if (!startMicrosecondsArray)
        {
            NSLog(@"%@: startMicrosecondsArray allocation of %u bytes failed", self, samples*sizeof(UnsignedWide) );
            [self release];
            return nil;
        }
        
        endMicrosecondsArray = calloc(samples,sizeof(UnsignedWide));
        if (!endMicrosecondsArray)
        {
            NSLog(@"%@: endMicrosecondsArray allocation of %u bytes failed", self, samples*sizeof(UnsignedWide) );
            [self release];
            return nil;
        }
        
        // we have buffers for our data, finish initializing
        
        outputData =            _outputData;
        outputResults =         _outputResults;
        disposeDataWhenDone=    _disposeDataWhenDone;
        nsLogStampStartEnd =    _nsLogStampStartEnd;
        unitMultiplier =        _multiplier;
        userLabel =             [_userLabel retain];
        // firstPass =             YES;
        // collectionNotComplete = YES;
        
        [self resetBenchmarker];
        
        // initialize unitsString
        
        if (unitMultiplier == millisecondsMultiplier)
        {
            unitsString = [@"milliseconds" retain];
        }
        else 
        {
            if (unitMultiplier == microsecondsMultiplier)
            {
                unitsString = [@"microseconds" retain];
            }
            else
            {
                if (unitMultiplier == nanosecondsMultiplier)
                {
                    unitsString = [@"nanoseconds" retain];
                }
                else
                {
                    if (unitMultiplier == noResultsRequested)
                    {
                        unitsString = [@"" retain];
                    }
                    else
                    {
                        unitsString = [@"unknown units" retain];
                    }
                }
            }
        }
        
        // we depend on the runtime to pre-initialize the instance variables to 0 for our
        // initialization of startBenchmarkIndex (0), endBenchmarkIndex (0), accumulatedTime (0), 
        // and dataDisposed (NO).
        
    }
    return self;
}

- (void)dealloc
{
    [self disposeData];
    [userLabel release];
    [unitsString release];
    
    [super dealloc];
}

- (void)disposeData
{
    if (startMicrosecondsArray) free(startMicrosecondsArray);
    if (endMicrosecondsArray) free(endMicrosecondsArray);
    startMicrosecondsArray = NULL;
    endMicrosecondsArray = NULL;
    dataDisposed = YES;
}


- (void)startBenchmark
{
    if (startBenchmarkIndex < samples)
    {
        if (firstPass)
        {
            firstPass = NO;
            if (nsLogStampStartEnd)
            {
                NSLog(@"%@ start benchmark", userLabel);
            }
        }
        Microseconds(&startMicrosecondsArray[startBenchmarkIndex]);
        startBenchmarkIndex++;
    }
}

- (void)endBenchmark
{
    if (endBenchmarkIndex < samples)
    {
        Microseconds(&endMicrosecondsArray[endBenchmarkIndex]);
        endBenchmarkIndex++;
        if (endBenchmarkIndex == samples)
        {
            collectionNotComplete = NO;     // data collection is complete
            if (nsLogStampStartEnd) NSLog(@"%@ end benchmark", userLabel);
            [self tallyData];
            if (outputResults) [self doOutputResults];
        }
    }
    
}

#pragma mark -

- (unsigned)sampleTag
{
    return startBenchmarkIndex;
}

- (void)endBenchmarkWithSampleTag:(unsigned)_tag
{
    Microseconds(&endMicrosecondsArray[_tag]);
    resultsReceivedCount++;
    if (resultsReceivedCount == samples)
    {
        collectionNotComplete = NO;     // data collection is complete
        if (nsLogStampStartEnd) NSLog(@"%@ end benchmark", userLabel);
        [self tallyData];
        if (outputResults) [self doOutputResults];
    }
}

#pragma mark -

- (void)tallyData
{
    const double millisecondsMultiplier = 0.001;
    // const double microsecondsMultiplier = 1.0;
    const double nanosecondsMultiplier = 1000.0;
    // const double noResultsRequested = 0.0;
 
    if (!dataDisposed)
    {
        unsigned benchmarkIndex;
        
        accumulatedTime = 0;
        for (benchmarkIndex = 0; benchmarkIndex < samples; benchmarkIndex++)
        {
            accumulatedTime += (endMicrosecondsArray[benchmarkIndex].lo - startMicrosecondsArray[benchmarkIndex].lo);
            if (outputData) printf("%u start, end:\t%u\t%u\n", 
                                                    benchmarkIndex, 
                        startMicrosecondsArray[benchmarkIndex].lo, 
                        endMicrosecondsArray[benchmarkIndex].lo); 
        }
        microseconds = ((double) accumulatedTime) / (double)samples;
        nanoseconds = microseconds * nanosecondsMultiplier;
        milliseconds = microseconds * millisecondsMultiplier;
    }
    else
    {
        NSLog(@"%@ tallyData after data disposed; results undefined.");
    }
}

- (void)doOutputResults
{
    printf("accumulatedTime = %g  %s for %u samples\n", ((double)accumulatedTime)*unitMultiplier, [unitsString UTF8String], samples);
    printf("benchmark average = %g %s\n", microseconds * unitMultiplier, [unitsString UTF8String]);
}

- (void)resetBenchmarker
{
    if (!dataDisposed)
    {
        nanoseconds = 0.0;
        microseconds = 0.0;
        milliseconds = 0.0;
        startBenchmarkIndex = 0;
        endBenchmarkIndex = 0;
        resultsReceivedCount = 0;
        firstPass = YES;
        collectionNotComplete = YES;
    }
    else
    {
        NSLog(@"%@ reset after data disposed; results undefined.");
    }
}

#pragma mark -

- (double)nanosecondsDoubleValue
{
    return nanoseconds;
}

- (float)nanosecondsFloatValue
{
    return (float)nanoseconds;
}

- (double)microsecondsDoubleValue
{
    return microseconds;
}

- (float)microsecondsFloatValue
{
    return (float)microseconds;
}

- (double)millisecondsDoubleValue
{
    return milliseconds;
}

- (float)millisecondsFloatValue
{
    return (float)milliseconds;
}

#pragma mark -

- (UnsignedWide *)startMicrosecondsArray
{
    return startMicrosecondsArray;
}

- (UnsignedWide *)endMicrosecondsArray
{
    return endMicrosecondsArray;
}

- (BOOL)getStartDatum:(UnsignedWide *)_microseconds atIdex:(unsigned)_index
{
    return NO;
}

- (BOOL)getEndDatum:(UnsignedWide *)_microseconds atIdex:(unsigned)_index
{
    return NO;
}

@end
